/****************************************************************************************
   Temperatur Logger - Software for logging temperature data with the DS18B20 and ESP32
   Written by Sebastian Hinz, https://www.diy-temperature-logger.com
   Licensed under GPLv3. See LICENSE for details.
****************************************************************************************/
#include <arduino.h>
#include "tempLoggerESP32.h"
#include "myNTPClient.h"

//////////////////////////////////////////////////////////////////////
MyNTPClient::MyNTPClient() : m_ntpClient(0), m_bValidTime(false), m_timeSummerWinter(false)
{

}

//////////////////////////////////////////////////////////////////////
MyNTPClient& MyNTPClient::GetInstance()
{
  static MyNTPClient instance;
  static bool bInit = false;
  if(!bInit)
  {
    // read configured settings 
    instance.setup();
  }

  // setup ntp client with configured time server
  static WiFiUDP  ntpUDP;
  static NTPClient ntpClient(ntpUDP, &instance.m_timeServer[0], 0, 60000);

  if(!bInit)
  {
    bInit = true;
    ntpClient.begin();
    if(ntpClient.forceUpdate())
    {
      instance.m_bValidTime = true;
    }

    instance.m_ntpClient = &ntpClient;
  }

  return instance;
}
//////////////////////////////////////////////////////////////////////
void MyNTPClient::setup()
{
  Basecamp& iot = MyBasecamp::GetInstance();
  MyBasecamp::getBasecampSemaphore();
  
  String key = String("Time_Server");
  String desc = String("Time Server, possible value are e.g. europe.pool.ntp.org. You can also use your router if it is configured as time server.");
  iot.web.addInterfaceElement(key, "input", desc, "#configform", key);
  
  // set defaults m_Name
  if (iot.configuration.get(key).length() == 0 ) 
  {
    iot.configuration.set(key, String("europe.pool.ntp.org"));
  }
  m_timeServer = iot.configuration.get(key);
  
  /////
  key = String("Time_ZoneOffset");
  desc = String("Offset to UTC. Set your time zone.");
  iot.web.addInterfaceElement(key, "input", desc, "#configform", key);
  iot.web.setInterfaceElementAttribute(key, "type", "number");
  iot.web.setInterfaceElementAttribute(key, "min", "-12");
  iot.web.setInterfaceElementAttribute(key, "max", "12");
  
  // set defaults Time_ZoneOffset
  if (iot.configuration.get(key).length() == 0 ) 
  {
    iot.configuration.set(key, String("0"));
  }
  m_timeZoneOffset = iot.configuration.get(key).toInt();


  /////
  key = String("Time_SummerWinter");
  desc = String("Enable summer / winter correction.");
  iot.web.addInterfaceElement(key, "input", desc, "#configform", key);
  iot.web.setInterfaceElementAttribute(key, "type", "number");
  iot.web.setInterfaceElementAttribute(key, "min", "0");
  iot.web.setInterfaceElementAttribute(key, "max", "1");

  // set defaults Time_ZoneOffset
  if (iot.configuration.get(key).length() == 0 ) 
  {
    iot.configuration.set(key, String("1"));
  }
  m_timeSummerWinter = iot.configuration.get(key).toInt();    


  MyBasecamp::releaseBasecampSemaphore();
  Serial.printf("Used TimeServer: %s, Time_ZoneOffset: %d\n", &m_timeServer[0], m_timeZoneOffset);
}

//////////////////////////////////////////////////////////////////////
bool MyNTPClient::summertime_EU(int year, byte month, byte day, byte hour, byte tzHours)
// European Daylight Savings Time calculation by "jurs" for German Arduino Forum
// input parameters: "normal time" for year, month, day, hour and tzHours (0=UTC, 1=MEZ)
// return value: returns true during Daylight Saving Time, false otherwise
{
 if (month<3 || month>10) return false; // keine Sommerzeit in Jan, Feb, Nov, Dez
 if (month>3 && month<10) return true; // Sommerzeit in Apr, Mai, Jun, Jul, Aug, Sep
 if (month==3 && (hour + 24 * day)>=(1 + tzHours + 24*(31 - (5 * year /4 + 4) % 7)) || month==10 && (hour + 24 * day)<(1 + tzHours + 24*(31 - (5 * year /4 + 1) % 7)))
   return true;
 else
   return false;
}

//////////////////////////////////////////////////////////////////////
void MyNTPClient::update()
{
  static int lastSecond= -1;
  static int lastHour = -1;

  if(m_ntpClient && m_ntpClient->update())
  {
    // if no valid initial time, update now
    m_bValidTime = true;
  }

  if (m_ntpClient && m_bValidTime)
  {
    DateTime dt(m_ntpClient->getEpochTime());

    // check every second
    if(dt.second() != lastSecond)
    {
      lastSecond = dt.second();
      //Serial.printf("Free Heap: %d\n", ESP.getFreeHeap());
    }
    
    // check every hour
    if(dt.hour() != lastHour)
    {
      lastHour = dt.hour();
  
      // correct summer/winter time
      if(m_timeSummerWinter && summertime_EU(dt.year(), dt.month(), dt.day(), dt.hour(), 0))
      {
        m_ntpClient->setTimeOffset(m_timeZoneOffset*3600+3600); // UTC <--> timezone summertime
      }
      else
      {
        m_ntpClient->setTimeOffset(m_timeZoneOffset*3600);  // UTC <--> timezone
      }


      // reset every day 0:00:00
      if(lastHour == 0)
      {
        Serial.printf("##########Restarting esp\n");
        //delay(2000);
       // ESP.restart();
      }
    }
  }

   
  /*
  NTPClient& ntpClient = MyNTPClient::GetInstance();

  long nowTime = ntpClient.getEpochTime();

  // check every 10 seconds
  if(fabs(nowTime - lastActionTime) > 10.0)
  {
    // correct summer/winter time
    lastActionTime = nowTime;
    ntpClient.update();
  
    DateTime dt(ntpClient.getEpochTime());
    if(summertime_EU(dt.year(), dt.month(), dt.day(), dt.hour(), 0))
    {
      ntpClient.setTimeOffset(7200); // UTC <--> MESZ
      //DateTime dt(ntpClient.getEpochTime());
    }
    else
    {
      ntpClient.setTimeOffset(3600);  // UTC <--> MEZ
      //DateTime dt(ntpClient.getEpochTime());
    }

    */
/*
    // reset if reset time is reached
    if((resetTime != 0) && (nowTime >= resetTime))
    {
      Serial.printf("##########Restarting esp %d >= %d\n", nowTime, resetTime);
      delay(3000);
      ESP.restart();
    }

    // sets the reset time to 0:00:00 next day 
    DateTime nowDT(nowTime);
//    DateTime newResetTime(nowDT.year(), nowDT.month(), nowDT.day(), 0, 0, 0); 
//    newResetTime = newResetTime + TimeSpan(1, 0, 0, 0); 
    DateTime newResetTime(nowDT.year(), nowDT.month(), nowDT.day(), nowDT.hour(), nowDT.minute(), 0); 
    newResetTime = newResetTime + TimeSpan(0, 0, 1, 0); 
    resetTime = newResetTime.unixtime();

    Serial.printf("newResetTime: %02d/%02d %02d:%02d:%02d, %d, %d\n", newResetTime.month(), newResetTime.day(), newResetTime.hour(), newResetTime.minute(), newResetTime.second(), nowTime, resetTime);

  }*/ 
} 

//////////////////////////////////////////////////////////////////////
// Code by JeeLabs http://news.jeelabs.org/code/
// Released to the public domain! Enjoy!
////////////////////////////////////////////////////////////////////////////////
// DateTime implementation - ignores time zones and DST changes
// NOTE: also ignores leap seconds, see http://en.wikipedia.org/wiki/Leap_second

const uint8_t daysInMonth [] PROGMEM = { 31,28,31,30,31,30,31,31,30,31,30,31 };

DateTime::DateTime (uint32_t t) 
{
  t -= SECONDS_FROM_1970_TO_2000;    // bring to 2000 timestamp from 1970

    ss = t % 60;
    t /= 60;
    mm = t % 60;
    t /= 60;
    hh = t % 24;
    uint16_t days = t / 24;
    uint8_t leap;
    for (yOff = 0; ; ++yOff) {
        leap = yOff % 4 == 0;
        if (days < 365 + leap)
            break;
        days -= 365 + leap;
    }
    for (m = 1; ; ++m) {
        uint8_t daysPerMonth = pgm_read_byte(daysInMonth + m - 1);
        if (leap && m == 2)
            ++daysPerMonth;
        if (days < daysPerMonth)
            break;
        days -= daysPerMonth;
    }
    d = days + 1;

    //Serial.printf("y=%d, m=%02d, d=%02d, H=%02d, m=%02d, s=%02d\n", yOff, m, d, hh, mm, ss);
}

